from .gcore.mutations import CreateAddress
from typing import List


class Receive:
    def __init__(self, cryptocurrency: str):
        self.cryptocurrency = cryptocurrency
    
    def create_address(self, response_fields: List):
        return CreateAddress().Mutate(
            cryptocurrency=self.cryptocurrency,
            response_fields=response_fields
        )



from .gcore.mutations import CreateDepositAccount
from typing import List, Optional
from .exc import InvalidClientObject


class NairaTokenAccount:
    def __init__(self, account_name):
        self.account_name = account_name
    
    def create_account(self, response_fields: List) -> str:
        return CreateDepositAccount().Mutate(
            accountName=self.accountName,
            response_fields=response_fields
            )

    def execute(self, client):
        try:
            return client.execute(query=self.create_account(
                response_fields=[
                    'accountNumber',
                    'accountName',
                    'accountType',
                    'bankName',
                    'accountReference'
                ]
            ))
        except AttributeError:
            raise InvalidClientObject("<BuyCoinsClient> object expected received {} instead".format(type(client)))
    


from .gcore.queries import GetNetworkFee, GetBalance
from .gcore.mutations import SendCoin
from typing import List, Optional
from .exc import SendLimitError, InvalidClientObject


class Send:
    def __init__(self, address: str, cryptocurrency: str, amount: float):
        self.address = address
        self.cryptocurrency = cryptocurrency
        self.amount = amount

    limits = {
        "bitcoin": 1,
        "ethereum": 50,
        "litecoin": 50,
        "nairatoken": 2000000
    }

    def execute(self, client):
        try:
            return client.execute(query=self.send())
        except AttributeError:
            raise InvalidClientObject("<BuyCoinsClient> object expected received {} instead".format(type(client)))

    def get_network_fee(self, response_fields):
        _price = GetNetworkFee()
        return _price.queryObject(
            response_fields=response_fields, 
            cryptocurrency=self.cryptocurrency, amount=self.amount
        )

    def check_limit(self):
        if Send.limits[self.cryptocurrency.lower()] < self.amount:
            return False
        else:
            return True

    def send(self, response_fields):
        if self.cryptocurrency.lower() in Send.limits.keys():
            if self.check_limit(self.amount, self.cryptocurrency):
                return SendCoin().Mutate(
                    cryptocurrency=self.cryptocurrency,
                    response_fields=response_fields,
                    amount=self.amount,
                    address=self.address
                )
            else:
                raise SendLimitError("Maximum daily transaction amount exceeded")

    def balance(self, response_fields: List):
        return GetBalance.queryObject(response_fields=response_fields)



from .gcore.queries import GetOrders, GetMarketBook, GetPrices
from .gcore.mutations import Buy, Sell
from typing import List, Optional
from .exc import InvalidClientObject


def get_prices(
    self, response_fields: List,
    cryptocurrency:  Optional[str]=None
) -> str:
    if cryptocurrency:
        return GetPrices().queryObject(
            cryptocurrency=cryptocurrency,
            response_fields=response_fields
        )
    else:
        return GetPrices().queryObject(
            response_fields=response_fields
        )


# Buying crytocurrency from Buycoins
class BuyCoins:
    def __init__(self, price: str, coin_amount: float, cryptocurrency):
        self.coin_amount = coin_amount
        self.cryptocurrency = cryptocurrency
        self.price = price

    def from_buycoins(self, response_fields) -> str:
        return Buy().Mutate(
            price=self.price,
            coin_amount=self.coin_amount,
            cryptocurrency=self.cryptocurrency,
            response_fields=response_fields
        )
    
    def execute(self, client):
        try:
            return client.execute(query=self.from_buycoins(
                response_fields=[
                    'id', 'cryptocurrency', 'status', 
                    'totalCoinAmount', 'side'
                ]
            ))
        except AttributeError:
            raise InvalidClientObject("<BuyCoinsClient> object expected received {} instead".format(type(client)))

class SellCoins:
    def __init__(self, price: str, coin_amount: float, cryptocurrency):
        self.coin_amount = coin_amount
        self.cryptocurrency = cryptocurrency
        self.price = price
    
    def sell(self, response_fields) -> str:
        return Sell().Mutate(
            coin_amount=self.coin_amount,
            price=self.price,
            response_fields=response_fields,
            cryptocurrency=self.cryptocurrency
        )

    def execute(self, client):
        try:
            return client.execute(query=self.sell())
        except AttributeError:
            raise InvalidClientObject("<BuyCoinsClient> object expected received {} instead".format(type(client)))



from typing import List, Optional


# GraphQL Base Query
class Query:
    def __init__(self):
        self.name = None

    def queryObject(self, response_fields: List, fields: Optional[List[tuple]]=None):
        if fields:
            mod_fields = [f"{i[0]}:{i[1]}" for i in fields]
            fields = tuple(i for i in mod_fields)
            newline = "\n                "
            result = f"""
            query {{
                {self.name}{f"{fields}"}{{
                    {newline.join(i for i in response_fields)}
                }}
            }}
            """
            return result.replace("\'", "")
        else:
            newline = "\n                    "
            result = f"""
            query {{
                {self.name}{{
                    {newline.join(i for i in response_fields)}
                }}
            }}
            """
            return result


# Get Sale Price
class BuyCoinsPrices(Query):
    """
    Buycoins prices
    """
    def __init__(self):
        super().__init__()
        self.name = "buycoinsPrices"


    def queryObject(self, side: str, mode: str, cryptocurrency: str, response_fields: List):
        newline = "\n                    "
        result = f"""
        query {{
            {self.name}(side: {side}, mode: {mode}, cryptocurrency: {cryptocurrency}){{
                {newline.join(i for i in response_fields)}
            }}
        }}
        """
        return result


class GetPrices(Query):
    def __init__(self):
        super().__init__()
        self.name = 'getPrices'

    def queryObject(self, response_fields: List, cryptocurrency: Optional[str]=None):
        if cryptocurrency:
            newline = "\n                    "
            result = f"""
            query {{
                {self.name} (cryptocurrency: {cryptocurrency}){{
                    {newline.join(i for i in response_fields)}
                }}
            }}
            """
            return result
        else:
            newline = "\n                    "
            result = f"""
            query {{
                {self.name}{{
                    {newline.join(i for i in response_fields)}
                }}
            }}
            """
            return result
class GetNetworkFee(Query):
    """
    Transfer charges
    """
    def __init__(self):
        super().__init__()
        self.name = "getEstimatedNetworkFee"
    
    def queryObject(self, cryptocurrency: str, amount: float, response_fields: List):
        newline = "\n                    "
        result = f"""
        query {{
            {self.name}(cryptocurrency: {cryptocurrency}, amount: {amount}) {{
                {newline.join(i for i in response_fields)}
            }}
        }}
        """
        return result


class GetOrders(Query):
    def __init__(self):
        super().__init__()
        self.name = "getOrders"
    
    def queryObject(self, status, response_fields: List):
        newline = "\n                    "
        result = f"""
        query {{
            {self.name}(status: {self.status}){{
                dynamicPriceExpiry
                orders {{
                    edges {{
                        node {{
                            {newline.join(i for i in response_fields)}
                        }}
                    }}
                }}
            }}
        }}
        """
        return result


class GetDynamicPriceExpiry(GetOrders):
    """
    Expiration date of a dynamic price
    """
    def __init__(self):
        self.name = 'getOrders'

    def queryObject(self, status):
        newline = "\n                    "
        result = f"""
        query {{
            {self.name}(status: {self.status}){{
                dynamicPriceExpiry
        }}
        """
        return result


class GetMarketBook(Query):
    def __init__(self):
        self.name = "getMarketBook"
    
    def queryObject(self, response_fields: List):
        newline = "\n                       "
        result = f"""
        query {{
            {self.name}{{
                dynamicPriceExpiry
                orders {{
                    edges {{
                        node {{
                            {newline.join(i for i in response_fields)}
                        }}
                    }}
                }}
            }}
        }}
        """
        return result


class GetBalance(Query):
    def __init__(self):
        super().__init__()
        self.name = "getBalances"



from .gcore.mutations import Buy, PostLimitOrder, PostMarketOrder
from .gcore.queries import GetOrders, GetMarketBook, GetDynamicPriceExpiry, BuyCoinsPrices
from typing import List, Optional
from .exc import InvalidClientObject


# P2P purchase
class BuyCoinsP2P:
    def __init__(
        self,
        order_side: str,
        coin_amount: float,
        cryptocurrency: str,
        price_type: str,
        status: str,
        static_price=None,
        dynamic_exchange_rate=None
    ):
        self.order_side = order_side
        self.coin_amount = coin_amount
        self.cryptocurrency = cryptocurrency
        self.price_type = price_type
        self.static_price = static_price
        self.dynamic_exchange_rate = dynamic_exchange_rate
        self.status = status

    def get_dynamic_price(self) -> str:
        return GetDynamicPriceExpiry().queryObject(
            status=self.status
        )

    def get_orders(self, response_fields: List) -> str:
        return GetOrders().queryObject(
            status=self.status,
            response_fields=response_fields
        )

    def get_market_book(self, response_fields: List) -> str:
        return GetMarketBook().queryObject(
            response_fields=response_fields
        )

    def post_limit_order(self, response_fields: List):
        return PostLimitOrder().Mutate(
            order_side=self.order_side,
            coin_amount=self.coin_amount,
            cryptocurrency=self.cryptocurrency,
            price_type=self.price_type,
            static_price=self.static_price,
            dynamic_exchange_rate=self.dynamic_exchange_rate,
            response_fields=response_fields
        )

    def post_market_order(self, response_fields: List):
        return PostMarketOrder().Mutate(
            order_side=self.order_side,
            coin_amount=self.coin_amount,
            cryptocurrency=self.cryptocurrency,
            response_fields=response_fields
        )

    def execute(self, client, limit_order=None):
        if limit_order:
            try:
                return client.execute(query=self.post_limit_order(response_fields=[
                    'id', 'minCoinAmount', 'cryptocurrency',
                    'maxBuy', 'status', 'minBuy', 'mode',
                    'minSell', 'maxSell', 'mode'
                    'buyPricePerCoin', 'sellPricePerCoin'
                ]))
            except AttributeError:
                raise InvalidClientObject("<BuyCoinsClient> object expected received {} instead".format(type(client)))
        else:
            try:
                return client.execute(query=self.post_market_order(response_fields=[
                    'id', 'coinAmount', 'cryptocurrency',
                    'side', 'status', 'createdAt',
                    'pricePerCoin', 'priceType',
                    'dynamicExchangeRate', 'staticPrice'
                ]))
            except AttributeError:
                raise InvalidClientObject("<BuyCoinsClient> object expected received {} instead".format(type(client)))
    


from .exc import MissingKeyError
from .sending import Send
from .gcore.queries import GetPrices
from .orders import BuyCoins, SellCoins
from .naira_token_account import NairaTokenAccount
from .receive import Receive
from base64 import b64encode
from typing import Optional, List
import requests


class BuyCoinsClient:
    """
    Executor class that handles all queries to the buycoins API.

    Args:
        public_key (str): generated public key from buycoins developer account

        secret_key (str): generated secret key from buycoins developer account
    
    Returns:
        <class 'BuycoinsClient'> object which can be used to execute requests.
    Raises:
        MissingKeyError: raised if either or none of the keys are provided when instantiating the class.
    """

    _URL = "https://backend.buycoins.tech/api/graphql"

    def __init__(self, public_key=None, secret_key=None):
        if not public_key:
            raise MissingKeyError("Missing Public_key")
        if not secret_key:
            raise MissingKeyError("Missing secret_key")
        api_key = b64encode(f"{public_key}:{secret_key}".encode('utf-8'))
        self.API_KEY = api_key.decode('utf-8')


    def set_headers(self):
        """
        sets headers for all HTTP requests sent by the client

        Args: None

        Returns: <class 'str'> representation of the header object.
        """
        return {
            "Authorization": f"Basic {self.API_KEY}"
        }

    def execute(self, query):
        """
        Executes any graphql query passed to it.

        Args:
            query (str): A string representation of the GraphQL query to be sent.
            This is usually generated by some helper classes and methods available in the library,
            however you may write raw queries in string format and pass them in. As an example:

            query = '''
                query{
                    getPrice(cryptocurrency: bitcoin)
                    {
                        id
                        side
                        status
                    }
                }
            '''
            resp = client.execute(query=query)
            print(resp)
        
        Returns: JSON formatted data from the buycoins server, but can be interfaced,
        as a dictionary <class 'Dict'> object. As such, with 'resp' from the previous
        example, we can do:
            print(resp['data'])
        """
        headers = self.set_headers()
        _req = requests.post(
            json={"query": query},
            url=BuyCoinsClient._URL,
            headers=headers
        )
         
        if _req.status_code == 200:
            return _req.json()
        else:
            print(_req)
    
    def get_active_price_id(
        self, response_fields: List,
        cryptocurrency=None
    ) -> str:
        """
        Helper method to fetch an active price ID for the purposes of making orders
        on the platform.

        Args:
            cryptocurrency (Optional[str]): a string value to indicate the crypto that's to
            be purchased, or that which its price ID is being sought after. 

        Returns: <class 'str'> an active ID if the cryptocurrency argument is specified
        otherwise it returns a generic list of dictionaries containing info about each 
        crypto-currency on the buycoins platform.
        """
        if cryptocurrency is None:
            query = GetPrices().queryObject(
                cryptocurrency=cryptocurrency
            )
            _req = requests.post(
                json={"query": query},
                url=BuyCoinsClient._URL,
                headers=self.set_headers()
            )
            if _req.status_code == 200:
                return _req.json()['data']['getPrices']
            else:
                return (_req.json(), _req.status_code)
        elif cryptocurrency is not None:
            query = GetPrices().queryObject(
                response_fields=response_fields
            )
            _req = requests.post(
                json={"query": query},
                url=BuyCoinsClient._URL,
                headers=self.set_headers()
            )
        if _req.status_code == 200:
            return _req.json()['data']['getPrices'][0]['id']
        else:
            return (_req.json(), _req.status_code)


client = BuyCoinsClient(public_key="I_8roV2FBaA",secret_key="n3n0CA3Zf3z1ADhAwUMv0CkeXt-xQqYP5Z31i0iGxA4")


from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='py-buycoins',
    packages=find_packages(),
    version='0.1.0',
    description='Python client library for BuyCoins Africa API',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Paul <paulcurious7@gmail.com>, Emmanuel Bashorun <bashorun.emma@gmail.com>',
    url='https://github.com/Bashorun97/BuyCoins-Python-SDK',
    classifiers=[
        # Project Maturity
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        "Operating System :: OS Independent",

        # License
        'License :: OSI Approved :: MIT License',

        # Python versions Support
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    python_requires='>=3'
)



class PycoinsException(Exception):
    """
    Buycoins python client error class
    """

    pass


class MissingKeyError(PycoinsException):
    pass


class SendLimitError(PycoinsException):
    pass


class InvalidClientObject(PycoinsException):
    pass



from typing import List, Optional


class Mutation:
    """
    The Mutation class is the parent class for all sub-mutation classes in the library.
    """
    def __init__(self):
        self.name = None

    def Mutate(self, fields: List[tuple], response_fields: List):
        """
        Automatically and crudely generates mutation queries on behalf of user.
        Should be used in place of the other subclasses, if user wishes to customize his/her
        queries completely from scratch.

        Args:
            fields (List[tuple]): A list of tuples with length of two(2).
            The first item (tuple[0]) in each tuple corresponds or should correspond to the name of a field to be specified
            as a mutation argument/field. The second item in each tuple (tuple[1]), corresponds to
            the value of tuple[0]. This would be such that:

                fields = [('name','Josh'),('age',14)]
            would correspond to:
                mutation{
                    createUser(name: 'Josh', age: 14){
                        response_fields
                    }
                }
            
            response_fields (List): A list of string values that correspond to the names of,
            response fields, that a user wants to be returned or specified in the response from
            the GraphQL API. Typically:
                subfields = ['id', 'name']
            would correspond to:
                mutation{
                    createUser(name: 'Josh', age: 14){
                        id,
                        name
                    }
                }
        """
        mod_fields = [f"{i[0]}: {str(i[1])}" for i in fields]
        _args = tuple(i.strip("\'") for i in mod_fields)
        newline = "\n                "
        result = f"""
        mutation {{
            {self.name}{_args} {{
                {newline.join(i for i in response_fields)}
            }}
        }}
        """
        return result.replace("\'", "")


class CreateDepositAccount(Mutation):
    def __init__(self):
        """
        Create deposit account mutation class inherits from <class 'Mutation'>
        """
        self.name = "createDepositAccount"
    
    def Mutate(self, accountName: str, response_fields: List):
        newline = "\n                "
        result = f"""
        mutation {{
            {self.name}(accountName: \"{f"{accountName}"}\") {{
                {newline.join(i for i in response_fields)}
            }}
        }}
        """
        return result


# Buy from BuyCoins
class Buy(Mutation):
    def __init__(self):
        """
        Buy crypto (from buycoins) mutation class inherits from <class 'Mutation'>
        """
        super().__init__()
        self.name = "buy"

    def Mutate(self, price: str, coin_amount: float, cryptocurrency, response_fields: List):
        newline = "\n                "
        result = f"""
        mutation {{
            {self.name}(price: \"{f"{price}"}\", coin_amount: {coin_amount}, cryptocurrency: {cryptocurrency})
            {{
                {newline.join(i for i in response_fields)}
            }}
        }}
        """
        return result


class PostLimitOrder(Mutation):
    """
    Post limit order mutation class inherits from <class 'Mutation'>
    """
    def __init__(self):
        super().__init__()
        self.name = "postLimitOrder"
        
    def Mutate(self, order_side: str, coin_amount: float, cryptocurrency: str, price_type: str, response_fields: List, static_price=None, dynamic_exchange_rate=None):
        newline = "\n                "
        if price_type == 'static':
            result = f"""
            mutation {{
                {self.name}(orderSide: {order_side}, coinAmount: {coin_amount}, cryptocurrency: {cryptocurrency}, staticPrice: {static_price}, priceType: {price_type})
                {{
                    {newline.join(i for i in response_fields)}
                }}
            }}
            """
        elif price_type == 'dynamic':
            result = f"""
            mutation {{
                {self.name}(orderSide: {order_side}, coinAmount: {coin_amount}, cryptocurrency: {cryptocurrency}, dynamicExchangeRate: {dynamic_exchange_rate}, priceType: {price_type})
                {{
                    {newline.join(i for i in response_fields)}
                }}
            }}
            """
        return result

    
class PostMarketOrder(Mutation):
    """
        Create deposit account mutation class inherits from <class 'Mutation'>
    """
    def __init__(self):
        super().__init__()
        self.name = "postMarketOrder"


    def Mutate(self, order_side: str, coin_amount: float, cryptocurrency: str, response_fields: List):
        newline = "\n                "
        result = f"""
        mutation {{
            {self.name}(orderSide: {order_side}, coinAmount: {coin_amount}, cryptocurrency: {cryptocurrency})
            {{
                {newline.join(i for i in response_fields)}
            }}
        }}
        """
        return result


class Sell(Mutation):
    def __init__(self):
        """
            Sell crypto mutation class inherits from <class 'Mutation'>
        """
        super().__init__()
        self.name = "sell"


    def Mutate(self, response_fields: List, price: str, coin_amount: float, cryptocurrency):
        newline = "\n                "
        result = f"""
        mutation {{
            {self.name}(price: \"{f"{price}"}\", coin_amount: {coin_amount}, cryptocurrency: {cryptocurrency})
            {{
                {newline.join(i for i in response_fields)}
            }}
        }}
        """
        return result


class SendCoin(Mutation):
    """
        Send crypto mutation class inherits from <class 'Mutation'>
    """
    def __init__(self):
        super().__init__()
        self.name = "send"

    def Mutate(self, cryptocurrency, address, amount, response_fields: List):
        newline = "\n                "
        result = f"""
        mutation {{
            {self.name}(cryptocurrency: {cryptocurrency}, amount: {amount}, address: \"{f"{address}"}\")
            {{
                {newline.join(i for i in response_fields)}
                transaction {{
                    hash
                    id
                }}
            }}
        }}
        """
        return result


class CreateAddress(Mutation):
    """
        Create address mutation class inherits from <class 'Mutation'>
    """
    def __init__(self):
        super().__init__()
        self.name = "createAddress"

    def Mutate(self, cryptocurrency, response_fields: List):
        newline = "\n                "
        result = f"""
        mutation {{
            {self.name}(cryptocurrency: {cryptocurrency})
            {{
                {newline.join(i for i in response_fields)}
            }}
        }}
        """
        return result
